--------------------------------------
-- 日本語文字列を入力するモジュール --
-- Programed by Ken Hashimoto       --
-- Version: 070417                  --
--------------------------------------

  dofile('putfont.lua')

function readint(fp, n)
  -- ファイルfpからnバイトの整数(BigEndian)を読み込む
  local rv = 0
  if fp == nil then return rv end
  for i = 1, n do
    local c
    c = fp:read(1)
    if c ~= nil then do
      rv = rv * 256 + string.byte(c)
      end
    end
  end
  return rv
end

function henkan(yomi)
  -- 単漢字辞書を検索し、yomiに対応する漢字と送り仮名(UTF-8)を返す
  -- 返り値は二次元配列で、添字は次の通り:
  --   [1][1]==漢字1 [1][2]==送り仮名1
  --   [2][1]==漢字2 [2][2]==送り仮名2
  --   [3][1]==漢字3 [3][2]==送り仮名2
  --   …
  local fp
  local rv = {}
  local n, yi
  local yn = 0   -- 読みの文字数
  local y8 = {}  -- 読み1字ずつ(UTF-8)
  local yu = {}  -- 読み1字ずつ(Unicode値)

  -- y8[], yu[], yn を設定
  while yomi ~= '' do
    yn = yn + 1
    y8[yn], yomi = utf8top(yomi)
    yu[yn] = utf82unicode(y8[yn])
  end
  if yn == 0 then return rv end

  -- 辞書ファイルオープン
  fp = io.open ('tankanji.dic', 'rb')
  if fp == nil then return rv end
  if fp:read(8) ~= 'TANKANJI' then do
    fp:close()
    return rv
    end
  end
  if fp:read(8) ~= '20070404' then do
    fp:close()
    return rv
    end
  end

  -- 最初の文字チェック
  while true do
    local c2 = readint(fp, 2)
    if c2 == 0 then return rv end
    if c2 == yu[1] then do
      local c3 = readint(fp, 3)
      fp:seek('set', c3)
      break
      end
    else do
      fp:seek('cur', 3)
      end
    end
  end

  -- メインループ
  yi = 1
  n = 0
  while true do
    -- 漢字と送り仮名をrv[]に格納
    while true do
      local c2, ka, okuri, i, flag
      c2 = readint(fp, 2)
      if c2 == 0 then break end
      ka = unicode2utf8(c2)
      flag = 0
      for j = 1, n do  -- 既に同じ漢字があったか確認
        if rv[j][1] == ka then do
          flag = 1
          break
          end
        end
      end
      if flag == 0 then do
        i = yi + 1
        okuri = ''
        while y8[i] ~= nil do
          okuri = okuri .. y8[i]
          i = i + 1
        end
        n = n + 1
        rv[n] = {}
        rv[n][1] = ka
        rv[n][2] = okuri
        end
      end
    end

    -- 次の文字をチェック
    yi = yi + 1
    while true do
      local c2 = readint(fp, 2)
      if c2 == 0 or yu[yi] == nil then do
        local flag
        fp:close()
        -- rv[] をソート
        flag = 1
        while flag > 0 do
          flag = 0
          for i = n - 1, 1, -1 do
            if string.len(rv[i][2]) > string.len(rv[i + 1][2]) then do
              local r1, r2
              r1 = rv[i][1]
              r2 = rv[i][2]
              rv[i][1] = rv[i + 1][1]
              rv[i][2] = rv[i + 1][2]
              rv[i + 1][1] = r1
              rv[i + 1][2] = r2
              flag = 1
              end
            end
          end
        end
        return rv
        end
      end
      if c2 == yu[yi] then do
        local c3 = readint(fp, 3)
        fp:seek('set', c3)
        break
        end
      else do
        fp:seek('cur', 3)
        end
      end
    end
  end
end

function putkouho(BG8Bit, x, y, ka)
  -- 候補として漢字1文字表示
  BG8Bit:Line(x, y, x + 15, y, 21)
  BG8Bit:Line(x, y + 15, x + 15, y + 15, 21)
  BG8Bit:Line(x, y + 1, x, y + 14, 21)
  BG8Bit:Line(x + 15, y + 1, x + 15, y + 14, 21)
  symbol(BG8Bit, x + 2, y + 2, 25, ka)
end

function jisho_henkan(BG8Bit, str, str0, kp, kouho)
  -- tankanji.dic による単漢字変換
  if kouho[1] == nil then do  -- 候補まだ出してない
    kouho = henkan(str0)
    if kouho[2] ~= nil then do  -- 複数の候補
      local y = 133
      local k = 1
      for j = 0, 2 do
        local x = 158
        for i = 0, 4 do
          if kouho[k] ~= nil then do
            -- 候補を1文字ずつ表示
            putkouho(BG8Bit, x, y, kouho[k][1])
            k = k + 1
            end
          end
          x = x + 19
        end
        y = y + 19
      end
      kp = 1
      end
    elseif kouho[1] ~= nil then do  -- 唯一つの候補
      local x = 1 + utf8len(str) * 6
      local str1 = kouho[1][1] .. kouho[1][2]
      box(BG8Bit, x, 0, x + utf8len(str0) * 6 - 1, 11, 0)
      symbol(BG8Bit, x, 0, 7, str1) -- 確定文字を白で表示
      str = str .. str1
      str0 = ''
      kouho = {}
      end
    end
    end
  else do  -- 候補が既に出ている
    if kouho[16] ~= nil then do
      local y = 133
      local k
      kp = kp + 15
      if kouho[kp] == nil then kp = 1 end
      k = kp
      for j = 0, 2 do
        local x = 158
        for i = 0, 4 do
          if kouho[k] ~= nil then do
            -- 候補を1文字ずつ表示
            box(BG8Bit, x + 1, y + 1, x + 14, y + 14, 0)
            putkouho(BG8Bit, x, y, kouho[k][1])
            k = k + 1
            end
          else box(BG8Bit, x, y, x + 15, y + 15, 0)
          end
          x = x + 19
        end
        y = y + 19
      end
      end
    end
    end
  end
  return str, str0, kp, kouho
end

function henkan_unicode(str)
  -- strの最初の4文字が全角16進ならそれを対応するUnicode文字
  -- に置き換える
  -- 次の場合は''を返却する
  --   最初の4文字が全角16進でない
  --   変換後の文字が半角英数or半角カナである
  --   変換後の文字がサロゲート文字である
  local u = 0       -- Unicode
  for i = 0, 3 do
    local keta = -1
    if string.find(str, "０") == 1 then keta =  0 end
    if string.find(str, "１") == 1 then keta =  1 end
    if string.find(str, "２") == 1 then keta =  2 end
    if string.find(str, "３") == 1 then keta =  3 end
    if string.find(str, "４") == 1 then keta =  4 end
    if string.find(str, "５") == 1 then keta =  5 end
    if string.find(str, "６") == 1 then keta =  6 end
    if string.find(str, "７") == 1 then keta =  7 end
    if string.find(str, "８") == 1 then keta =  8 end
    if string.find(str, "９") == 1 then keta =  9 end
    if string.find(str, "Ａ") == 1 then keta = 10 end
    if string.find(str, "ａ") == 1 then keta = 10 end
    if string.find(str, "Ｂ") == 1 then keta = 11 end
    if string.find(str, "ｂ") == 1 then keta = 11 end
    if string.find(str, "Ｃ") == 1 then keta = 12 end
    if string.find(str, "ｃ") == 1 then keta = 12 end
    if string.find(str, "Ｄ") == 1 then keta = 13 end
    if string.find(str, "ｄ") == 1 then keta = 13 end
    if string.find(str, "Ｅ") == 1 then keta = 14 end
    if string.find(str, "ｅ") == 1 then keta = 14 end
    if string.find(str, "Ｆ") == 1 then keta = 15 end
    if string.find(str, "ｆ") == 1 then keta = 15 end
    if keta == -1 then return '' end  -- 16進文字ではなかった
    u = u * 16 + keta
    str = string.sub(str, 4)
  end
  if u < 128 or (65376 <= u and u < 65440) then return '' end  -- 半角文字
  if u >= 55296 and u <= 57343 then return '' end  -- サロゲート文字
  return unicode2utf8(u) .. str
end

function utf8add(str, kc)
  -- UTF-8文字列strにキーkcを全角文字として追加する
  if kc == 222 then do -- 濁点
    if string.sub(str, -3) == 'か' then do
      str = string.sub(str, 1, -4) .. 'が' end
    elseif string.sub(str, -3) == 'き' then do
      str = string.sub(str, 1, -4) .. 'ぎ' end
    elseif string.sub(str, -3) == 'く' then do
      str = string.sub(str, 1, -4) .. 'ぐ' end
    elseif string.sub(str, -3) == 'け' then do
      str = string.sub(str, 1, -4) .. 'げ' end
    elseif string.sub(str, -3) == 'こ' then do
      str = string.sub(str, 1, -4) .. 'ご' end
    elseif string.sub(str, -3) == 'さ' then do
      str = string.sub(str, 1, -4) .. 'ざ' end
    elseif string.sub(str, -3) == 'し' then do
      str = string.sub(str, 1, -4) .. 'じ' end
    elseif string.sub(str, -3) == 'す' then do
      str = string.sub(str, 1, -4) .. 'ず' end
    elseif string.sub(str, -3) == 'せ' then do
      str = string.sub(str, 1, -4) .. 'ぜ' end
    elseif string.sub(str, -3) == 'そ' then do
      str = string.sub(str, 1, -4) .. 'ぞ' end
    elseif string.sub(str, -3) == 'た' then do
      str = string.sub(str, 1, -4) .. 'だ' end
    elseif string.sub(str, -3) == 'ち' then do
      str = string.sub(str, 1, -4) .. 'ぢ' end
    elseif string.sub(str, -3) == 'つ' then do
      str = string.sub(str, 1, -4) .. 'づ' end
    elseif string.sub(str, -3) == 'て' then do
      str = string.sub(str, 1, -4) .. 'で' end
    elseif string.sub(str, -3) == 'と' then do
      str = string.sub(str, 1, -4) .. 'ど' end
    elseif string.sub(str, -3) == 'は' then do
      str = string.sub(str, 1, -4) .. 'ば' end
    elseif string.sub(str, -3) == 'ひ' then do
      str = string.sub(str, 1, -4) .. 'び' end
    elseif string.sub(str, -3) == 'ふ' then do
      str = string.sub(str, 1, -4) .. 'ぶ' end
    elseif string.sub(str, -3) == 'へ' then do
      str = string.sub(str, 1, -4) .. 'べ' end
    elseif string.sub(str, -3) == 'ほ' then do
      str = string.sub(str, 1, -4) .. 'ぼ' end
    else do
      str = str .. '゛' end
    end
    end
  elseif kc == 223 then do -- 半濁点
    if string.sub(str, -3) == 'は' then do
      str = string.sub(str, 1, -4) .. 'ぱ' end
    elseif string.sub(str, -3) == 'ひ' then do
      str = string.sub(str, 1, -4) .. 'ぴ' end
    elseif string.sub(str, -3) == 'ふ' then do
      str = string.sub(str, 1, -4) .. 'ぷ' end
    elseif string.sub(str, -3) == 'へ' then do
      str = string.sub(str, 1, -4) .. 'ぺ' end
    elseif string.sub(str, -3) == 'ほ' then do
      str = string.sub(str, 1, -4) .. 'ぽ' end
    else do
      str = str .. '゜' end
    end
    end
  else do -- 普通の文字
    if kc >= 161 then kc = kc + 65216 end
    kc = han2zen(kc)
    if kc ~= nil then do
      if kc >= 12449 and kc <= 12531 then kc = kc - 96 end -- ひらがなにする
      str = str .. unicode2utf8(kc)
      end
    end
    end
  end
  return str
end

function hzlamp(BG8Bit, mode_hz)
  -- mode_hzに従ってインジケータ表示
  box(BG8Bit, 153, 54, 155, 60, (1 - mode_hz) * 2)
  box(BG8Bit, 205, 54, 207, 60, mode_hz * 2)
end

function putcursor(BG8Bit, len)
  -- lenの位置にカーソル表示
  local x = len * 6 + 1
  box(BG8Bit, x, 0, x + 1, 11, 6)
end

function erasecursor(BG8Bit, len)
  -- lenの位置のカーソルを消去
  local x = len * 6 + 1
  box(BG8Bit, x, 0, x + 1, 11, 0)
end

function keycode()
  -- ペンで押されているソフトキーのコードを返却する
  --   何も押されていない: 0
  --   通常キー: 32～126、161～223
  --   BackSpace: 7
  --   Enter: 13
  --   変換候補: 300～314
  --   半角: 350
  --   全角: 351
  --   ひらがな確定: 352
  --   カタカナ確定: 353
  --   変換/次候補: 355

  local x, y

  if Stylus.Down() == false then return 0 end

  x = Stylus.X()
  y = Stylus.Y()

  -- 英数文字キー
  if x <= 87 then do
    local x0 = -1
    local y0 = -1
    for i = 0, 5 do
      local x1
      x1 = 5 + i * 14
      if x1 <= x and x <= x1 + 12 then x0 = i end
    end
    for i = 0, 15 do
      local y1
      y1 = 15 + i * 11
      if y1 <= y and y <= y1 + 9 then y0 = i end
    end
    if x0 == -1 or y0 == -1 then return 0 end
    if x0 == 5 and y0 == 15 then return 0 end
    return 32 + x0 * 16 + y0
    end
  end

  -- カナ文字キー
  if x <= 146 then do
    local x0 = -1
    local y0 = -1
    for i = 0, 3 do
      local x1
      x1 = 92 + i * 14
      if x1 <= x and x <= x1 + 12 then x0 = i end
    end
    for i = 0, 15 do
      local y1
      y1 = 15 + i * 11
      if y1 <= y and y <= y1 + 9 then y0 = i end
    end
    if x0 == -1 or y0 == -1 then return 0 end
    if x0 == 0 and y0 == 0 then return 0 end
    return 160 + x0 * 16 + y0
    end
  end

  -- 変換候補欄
  if y >= 133 then do
    local x0 = -1
    local y0 = -1
    for i = 0, 4 do
      local x1
      x1 = 159 + i * 19
      if x1 <= x and x <= x1 + 13 then x0 = i end
    end
    for i = 0, 2 do
      local y1
      y1 = 134 + i * 19
      if y1 <= y and y <= y1 + 13 then y0 = i end
    end
    if x0 == -1 or y0 == -1 then return 0 end
    return 300 + y0 * 5 + x0
    end
  end

  -- BackSpace、Enterキー
  if 17 <= y and y <= 39 then do
    if 159 <= x and x <= 197 then return 7 end
    if 211 <= x and x <= 249 then return 13 end
    return 0
    end
  end

  -- 半角、全角キー
  if 46 <= y and y <= 68 then do
    if 159 <= x and x <= 197 then return 350 end
    if 211 <= x and x <= 249 then return 351 end
    return 0
    end
  end

  -- 無変換キー
  if 75 <= y and y <= 97 then do
    if 159 <= x and x <= 197 then return 352 end
    if 211 <= x and x <= 249 then return 353 end
    return 0
    end
  end

  -- 変換/次候補キー
  if 104 <= y and y <= 126 then do
    if 211 <= x and x <= 249 then return 355 end
    return 0
    end
  end

  return 0
end

function kinput(BG8Bit, palnum, str)
  -- 日本語文字列をタッチペンで入力する
  --   BG8Bit : 下画面として初期化されたオブジェクト
  --   palnum : ソフトキー用パレット番号
  --   str    : デフォルト文字列(UTF-8である事)
  --   返り値 : 入力された文字列(UTF-8)

  local KbdFr, Kbd00, Kbd01, Kbd02, Kbd10, Kbd11, Kbd12
  local Kbd20, Kbd21, Kbd22, Kbd30, Kbd31
  local kcold = 0
  local kc = 0
  local mode_hz = 0     -- 0:半角モード 1:全角モード
  local str0            -- 未確定文字列
  local kouho = {}      -- 変換候補
  local kp              -- 候補表示開始オフセット(1,16,31,…)

  -- ソフトキーボードのスプライトを表示
  Sprite.LoadPalette(0, palnum, 'kbd_x64.pal')
  KbdFr = FrameStrip.Create(0, 64, 64, 256)
  KbdFr:LoadBin('kbd_x64.raw', 11) 
  Kbd00 = Sprite.Create(KbdFr,  0, palnum,   0,   0)
  Kbd01 = Sprite.Create(KbdFr,  1, palnum,   0,  64)
  Kbd02 = Sprite.Create(KbdFr,  2, palnum,   0, 128)
  Kbd10 = Sprite.Create(KbdFr,  3, palnum,  64,   0)
  Kbd11 = Sprite.Create(KbdFr,  4, palnum,  64,  64)
  Kbd12 = Sprite.Create(KbdFr,  5, palnum,  64, 128)
  Kbd20 = Sprite.Create(KbdFr,  6, palnum, 128,   0)
  Kbd21 = Sprite.Create(KbdFr,  7, palnum, 128,  64)
  Kbd22 = Sprite.Create(KbdFr,  8, palnum, 128, 128)
  Kbd30 = Sprite.Create(KbdFr,  9, palnum, 192,   0)
  Kbd31 = Sprite.Create(KbdFr, 10, palnum, 192,  64)

  -- パレット設定
  -- 通常
  BG8Bit:SetPaletteColor( 0,  0,  0,  0)
  BG8Bit:SetPaletteColor( 1,  0,  0, 31)
  BG8Bit:SetPaletteColor( 2, 31,  0,  0)
  BG8Bit:SetPaletteColor( 3, 31,  0, 31)
  BG8Bit:SetPaletteColor( 4,  0, 31,  0)
  BG8Bit:SetPaletteColor( 5,  0, 31, 31)
  BG8Bit:SetPaletteColor( 6, 31, 31,  0)
  BG8Bit:SetPaletteColor( 7, 31, 31, 31)
  -- 暗い
  BG8Bit:SetPaletteColor(11,  0,  0, 15)
  BG8Bit:SetPaletteColor(12, 15,  0,  0)
  BG8Bit:SetPaletteColor(13, 15,  0, 15)
  BG8Bit:SetPaletteColor(14,  0, 15,  0)
  BG8Bit:SetPaletteColor(15,  0, 15, 15)
  BG8Bit:SetPaletteColor(16, 15, 15,  0)
  BG8Bit:SetPaletteColor(17, 15, 15, 15)
  -- 明るい
  BG8Bit:SetPaletteColor(20, 15, 15, 15)
  BG8Bit:SetPaletteColor(21, 15, 15, 31)
  BG8Bit:SetPaletteColor(22, 31, 15, 15)
  BG8Bit:SetPaletteColor(23, 31, 15, 31)
  BG8Bit:SetPaletteColor(24, 15, 31, 15)
  BG8Bit:SetPaletteColor(25, 15, 31, 31)
  BG8Bit:SetPaletteColor(26, 31, 31, 15)

  -- グラフィック画面消去
  box(BG8Bit, 0, 0, 255, 191, 0)

  -- ソフトキーボード点灯
  box(BG8Bit, 4, 14, 147, 190, 25)  -- 通常キーの文字
  for x = 4, 88, 14 do              -- 通常キーの縦線
    BG8Bit:Line(x, 14, x, 190, 15)
  end
  for x = 91, 147, 14 do            -- 通常キーの縦線
    BG8Bit:Line(x, 14, x, 190, 15)
  end
  for y = 14, 190, 11 do            -- 通常キーの横線
    BG8Bit:Line(4, y, 147, y, 15)
  end
  box(BG8Bit, 209, 15, 251, 41, 26) -- Enter

  -- デフォルト文字列から不正な文字を除去
  str0 = ''
  while str ~= '' do
    local u
    u, str = utf82unicode(str)
    if u > 65535 then u = -1 end
    if u >= 55296 and u <= 57343 then u = -1 end -- サロゲート文字
    if u >= 0 then str0 = str0 .. unicode2utf8(u) end
  end
  str = str0
  str0 = ''

  -- デフォルト文字列表示
  while utf8len(str) > 42 do str = utf8chop(str) end
  symbol(BG8Bit, 1, 0, 7, str)

  -- メインループ
  while true do
    kc = keycode()
    if kc ~= 0 and kc ~= kcold then do  -- 入力あり
      erasecursor(BG8Bit, utf8len(str) + utf8len(str0)) -- カーソル消去
      if kc == 7 then do
        -- BackSpace
        if str0 == '' then do
          -- 未確定文字列無し
          local olen, nlen, x
          olen = utf8len(str)
          str = utf8chop(str)
          nlen = utf8len(str)
          box(BG8Bit, nlen * 6 + 1, 0, olen * 6, 11, 0) -- 1文字消去
          end
        else do
          -- 未確定文字列有り
          if kouho[1] == nil then do  -- 変換候補無し
            local blen, olen, nlen, x
            blen = utf8len(str)
            olen = utf8len(str0)
            str0 = utf8chop(str0)
            nlen = utf8len(str0)
            -- 1文字消去
            box(BG8Bit, (blen + nlen) * 6 + 1, 0, (blen + olen) * 6, 11, 0)
            end
          elseif kp == 1 then do  -- 変換候補の先頭ページ
            -- 候補を消す
            box(BG8Bit, 158, 133, 249, 186, 0)
            kouho = {}
            end
          else do  -- 変換候補の2ページ目以降
            -- 前候補へ
            local y = 133
            local k
            kp = kp - 15
            k = kp
            for j = 0, 2 do
              local x = 158
              for i = 0, 4 do
                -- 候補を1文字ずつ表示
                box(BG8Bit, x + 1, y + 1, x + 14, y + 14, 0)
                putkouho(BG8Bit, x, y, kouho[k][1])
                k = k + 1
                x = x + 19
              end
              y = y + 19
            end
            end
          end
          end
        end
        end
      elseif kc >= 32 and kc <= 223 then do
        -- 通常キー
        if mode_hz == 0 then do
          -- 半角モード
          if utf8len(str) < 42 then do
            -- 32～126、161～223
            local c
            if kc < 128 then do
              c = string.char(kc) -- 半角英数
              end
            elseif kc <= 191 then do
              c = string.char(239, 189, kc) -- 半角カナ(UTF-8)
              end
            else do
              c = string.char(239, 190, kc - 64) -- 半角カナ(UTF-8)
              end
            end
            symbol(BG8Bit, 1 + utf8len(str) * 6, 0, 7, c) -- 表示
            str = str .. c
            end
          end
          end
        else do
          -- 全角モード
          if str0 == '' and kc == 32 then do
            str = utf8add(str, kc)
            if utf8len(str) > 42 then str = utf8chop(str) end
            end
          else do
            str0 = utf8add(str0, kc)
            if kouho[1] ~= nil then do
              box(BG8Bit, 158, 133, 249, 186, 0)
              kouho = {}
              end
            end
            if utf8len(str) + utf8len(str0) > 42 then do
              str0 = utf8chop(str0)
              end
            else do
              local x
              local str1 = utf8last(str0)
              x = 1 + (utf8len(str) + utf8len(utf8chop(str0))) * 6
              box(BG8Bit, x, 0, x + 11, 11, 0)
              symbol(BG8Bit, x, 0, 4, str1) -- 未確定文字として表示
              end
            end
            end
          end
          end
        end
        end
      elseif kc == 350 then do
        -- 半角キー
        local x = 1 + utf8len(str) * 6
        if kouho[1] ~= nil then do
          box(BG8Bit, 158, 133, 249, 186, 0)
          kouho = {}
          end
        end
        symbol(BG8Bit, x, 0, 7, str0) -- 未確定文字を白で表示
        str = str .. str0
        mode_hz = 0
        str0 = ''
        end
      elseif kc == 351 then do
        -- 全角キー
        mode_hz = 1
        end
      elseif kc == 352 then do
        -- ひらがな確定
        local x = 1 + utf8len(str) * 6
        if kouho[1] ~= nil then do
          box(BG8Bit, 158, 133, 249, 186, 0)
          kouho = {}
          end
        end
        symbol(BG8Bit, x, 0, 7, str0) -- 確定文字を白で表示
        str = str .. str0
        str0 = ''
        end
      elseif kc == 353 then do
        -- カタカナ確定
        local str1, str2
        local x = 1 + utf8len(str) * 6
        str1 = str0
        str2 = '' -- 変換後文字列
        while str1 ~= '' do
          local u
          u, str1 = utf82unicode(str1)
          if u >= 12353 and u <= 12438 then u = u + 96 end
          str2 = str2 .. unicode2utf8(u)
        end
        str2 = string.gsub(str2, "ウ゛", "ヴ")
        if kouho[1] ~= nil then do
          box(BG8Bit, 158, 133, 249, 186, 0)
          kouho = {}
          end
        end
        box(BG8Bit, x, 0, x + utf8len(str0) * 6 - 1, 11, 0)
        symbol(BG8Bit, x, 0, 7, str2) -- 確定文字を白で表示
        str = str .. str2
        str0 = ''
        end
      elseif kc == 355 then do
        -- 変換/次候補
        if str0 ~= '' then do
          local str1 = henkan_unicode(str0)
          if str1 ~= '' then do
            local x = 1 + utf8len(str) * 6
            box(BG8Bit, x, 0, x + utf8len(str0) * 6 - 1, 11, 0)
            symbol(BG8Bit, x, 0, 7, str1) -- 確定文字を白で表示
            str = str .. str1
            str0 = ''
            end
          else do
            str, str0, kp, kouho = jisho_henkan(BG8Bit, str, str0, kp, kouho)
            end
          end
          end
        end
        end
      elseif kc >= 300 and kc <= 314 then do
        -- 変換候補選択
        if kouho[1] ~= nil then do
          local kn = kc - 300 + kp
          if kouho[kn] ~= nil then do
            local str1 = kouho[kn][1] .. kouho[kn][2]
            local x = 1 + utf8len(str) * 6
            box(BG8Bit, x, 0, x + utf8len(str0) * 6 - 1, 11, 0)
            symbol(BG8Bit, x, 0, 7, str1) -- 確定文字を白で表示
            str = str .. str1
            box(BG8Bit, 158, 133, 249, 186, 0)
            str0 = ''
            kouho = {}
            end
          end
          end
        end
        end
      else do
        -- Enter ( kc == 13 )
        if str0 == '' then do
          -- 未確定文字列無し
          break
          end
        else do
          -- 未確定文字列有り
          local x = 1 + utf8len(str) * 6
          symbol(BG8Bit, x, 0, 7, str0) -- 未確定文字を白で表示
          str = str .. str0
          if kouho[1] ~= nil then box(BG8Bit, 158, 133, 249, 186, 0) end
          break
          end
        end
        end
      end
      end
    else do                 -- 入力なし
      local kclr -- 特殊キーの色
      if str ~= '' or str0 ~= '' then kclr = 26 else kclr = 17 end
      box(BG8Bit, 157, 15, 199, 41, kclr)  -- BS
      if mode_hz == 1 or str0 ~= '' then kclr = 26 else kclr = 17 end
      box(BG8Bit, 157, 44, 199, 70, kclr)  -- 半角
      if mode_hz == 0 then kclr = 26 else kclr = 17 end
      box(BG8Bit, 209, 44, 251, 70, kclr)  -- 全角
      if str0 ~= '' then kclr = 26 else kclr = 17 end
      box(BG8Bit, 157, 73, 199, 99, kclr)  -- ひら確定
      box(BG8Bit, 209, 73, 251, 99, kclr)  -- カタ確定
      box(BG8Bit, 209, 102, 251, 128, kclr)  -- 変換/次候補
      hzlamp(BG8Bit, mode_hz) -- 半角全角モードインジケート
      putcursor(BG8Bit, utf8len(str) + utf8len(str0)) -- カーソル表示
      end
    end
    kcold = kc
  end

  -- スプライトを解放
  Kbd00:Free()
  Kbd01:SetFrame(KbdFr, 0)
  Kbd01:Free()
  Kbd02:SetFrame(KbdFr, 0)
  Kbd02:Free()
  Kbd10:SetFrame(KbdFr, 0)
  Kbd10:Free()
  Kbd11:SetFrame(KbdFr, 0)
  Kbd11:Free()
  Kbd12:SetFrame(KbdFr, 0)
  Kbd12:Free()
  Kbd20:SetFrame(KbdFr, 0)
  Kbd20:Free()
  Kbd21:SetFrame(KbdFr, 0)
  Kbd21:Free()
  Kbd22:SetFrame(KbdFr, 0)
  Kbd22:Free()
  Kbd30:SetFrame(KbdFr, 0)
  Kbd30:Free()
  Kbd31:SetFrame(KbdFr, 0)
  Kbd31:Free()

  -- 画面消去
  box(BG8Bit, 0, 0, 255, 191, 0)

  return str
end
